﻿using System.Diagnostics;
using System.IO;
using System.Management;
using System.Windows;
using Newtonsoft.Json;

namespace InnoLabProjektDektopApp.Services
{
    public class ProcessInfo(string processName, DateTime startTime, DateTime endTime)
    {
        public string ProcessName { get; set; } = processName;
        public DateTime StartTime { get; set; } = startTime;
        public DateTime EndTime { get; set; } = endTime;
    }


    class ProcessMonitor
    {
        private readonly Dictionary<string, DateTime> processStartTimes = new Dictionary<string, DateTime>();
        private static List<string> blockedProcesses = new List<string>();
        private bool isMonitoring = false;
        private readonly List<ProcessInfo> processInfoList = new List<ProcessInfo>();
        private DateTime startTime;

        public ProcessMonitor()
        {
            blockedProcesses = LoadBlockedProcesses();

        }

        public async void StartMonitoring()
        {
            startTime = DateTime.Now;
            Debug.WriteLine("Started Process Monitoring");
            var query = new WqlEventQuery("__InstanceCreationEvent", new TimeSpan(0, 0, 1),
                "TargetInstance isa \"Win32_Process\"");
            var watcher = new ManagementEventWatcher { Query = query };

            isMonitoring = true;
            await Task.Run(() =>
            {
                while (isMonitoring)
                {
                    // this wait is here to prevent another event from being triggered before the previous one is handled
                    var e = watcher.WaitForNextEvent();
                    Task.Run(() => HandleNewProcess(e));
                    Task.Delay(100).Wait();
                }
            });
        }

        public void StopMonitoring()
        {
            Debug.WriteLine("Stopping Process Monitoring");
            CalculateCurrentDistractionStage(DateTime.Now);
            isMonitoring = false;
        }

        public void CalculateCurrentDistractionStage(DateTime endTime)
        {
            TimeSpan totalDistractionTime = CalculateTotalDistractionTime();
            Debug.WriteLine("Total distraction time: {0}", totalDistractionTime);
            double totalMinutes = (endTime - startTime).TotalMinutes;
            double distractedMinutes = totalDistractionTime.TotalMinutes;
            double distractionFraction = (distractedMinutes / totalMinutes);
            int stage = (int)Math.Ceiling(distractionFraction * 7);
            Debug.WriteLine($"Distraction stage: {stage}");
        }

        private void HandleNewProcess(ManagementBaseObject e)
        {
            var processName = Path.GetFileNameWithoutExtension((string)((ManagementBaseObject)e["TargetInstance"])["Name"]);

            if (processStartTimes.TryGetValue(processName, out _))
            {
                return;
            }

            processStartTimes[processName] = DateTime.Now;

            if (IsProcessOnBlockedList(processName))
            {
                var result = MessageBox.Show($"Are you sure you want to get distracted by {processName}?", "Distracting Process Detected",
                MessageBoxButton.YesNo,
                MessageBoxImage.Question,
                MessageBoxResult.No,
                MessageBoxOptions.DefaultDesktopOnly
                );

                if (result == MessageBoxResult.No)
                {
                    KillProcessesByName(processName);
                    processStartTimes.Remove(processName);
                }
                else
                {
                    TrackProcess(processName, DateTime.Now);
                }
            }
        }

        private static void KillProcessesByName(String processName)
        {
            var runningProcesses = GetRunningProcessList();
            if (!runningProcesses.TryGetValue(processName, out List<Process>? processes) || processes == null)
            {
                return; // Return early if the process name is not found or processes is null
            }

            Debug.WriteLine("Killing {0} processes with the Name {1}", processes.Count, processName);
            foreach (var process in processes)
            {
                process.Kill();
            }
        }

        private static Dictionary<string, List<Process>> GetRunningProcessList()
        {
            var processlist = Process.GetProcesses();
            var processDictionary = new Dictionary<string, List<Process>>();

            foreach (var process in processlist)
            {
                var processName = process.ProcessName;

                if (!processDictionary.TryGetValue(processName, out List<Process>? value))
                {
                    value = [];
                    processDictionary[processName] = value;
                }

                value.Add(process);
            }

            return processDictionary;
        }

        private static bool IsProcessOnBlockedList(string processName)
        {
            return blockedProcesses.Contains(processName);
        }

        private async void TrackProcess(string processName, DateTime startTime)
        {
            Debug.WriteLine($"Start monitoring the {processName} process");

            Process? process = GetRunningProcessList().TryGetValue(processName, out List<Process>? processes) ? processes[0] : null;

            if (process == null)
            {
                Debug.WriteLine("Process exited before we could track it.");
                return;
            }

            // Token to handle task cancellation
            using var cts = new CancellationTokenSource();
            var token = cts.Token;

            // Task to wait for the process to exit
            var waitForExitTask = Task.Run(() =>
            {
                process.WaitForExit();
                SaveTrackedData(processName, startTime);
            }, token);

            // Task to monitor the isMonitoring flag
            var monitorTask = Task.Run(async () =>
            {
                while (isMonitoring)
                {
                    await Task.Delay(100); // Check every 100ms
                }
            }, token);

            try
            {
                // Await the completion of either the waitForExitTask or the monitorTask
                await Task.WhenAny(waitForExitTask, monitorTask);
                cts.Cancel();
            }
            catch (OperationCanceledException)
            {
                Debug.WriteLine("Process monitoring was canceled.");
            }
        }

        private void SaveTrackedData(String processName, DateTime startTime)
        {
            processStartTimes.Remove(processName);
            var endTime = DateTime.Now;
            processInfoList.Add(new ProcessInfo(processName, startTime, endTime));
            Debug.WriteLine("Process monitoring complete.");
            Debug.WriteLine("The Process has been running for {0} seconds", endTime - startTime);
        }


        public TimeSpan CalculateTotalDistractionTime()
        {
            var tempProcessInfoList = new List<ProcessInfo>();

            // Add the earliest start time entry from processStartTimes to tempProcessInfoList
            if (processStartTimes.Count > 0)
            {
                var earliestProcess = processStartTimes.OrderBy(p => p.Value).First();
                tempProcessInfoList.Add(new ProcessInfo(earliestProcess.Key, earliestProcess.Value, DateTime.Now));
            }

            // Add all entries from processInfoList to tempProcessInfoList
            tempProcessInfoList.AddRange(processInfoList);

            var intervals = tempProcessInfoList
                .OrderBy(p => p.StartTime)
                .Select(p => (p.StartTime, p.EndTime))
                .ToList();

            if (intervals.Count == 0)
                return TimeSpan.Zero;

            var mergedIntervals = new List<(DateTime Start, DateTime End)> { intervals[0] };

            foreach (var interval in intervals.Skip(1))
            {
                var last = mergedIntervals.Last();
                if (interval.StartTime <= last.End)
                {
                    mergedIntervals[^1] = (last.Start, new DateTime(Math.Max(last.End.Ticks, interval.EndTime.Ticks)));
                }
                else
                {
                    mergedIntervals.Add(interval);
                }
            }

            return mergedIntervals.Aggregate(TimeSpan.Zero, (sum, interval) => sum + (interval.End - interval.Start));
        }


        private static List<string> LoadBlockedProcesses()
        {
            var blockedProcessesFilePath = Path.Combine(AppDomain.CurrentDomain.BaseDirectory, "Assets\\blockedProcesses.json");
            var fileContent = File.ReadAllText(blockedProcessesFilePath);
            return JsonConvert.DeserializeObject<List<string>>(fileContent) ?? [];
        }
    }
}
